/*
 * Copyright (C) 1997-2004 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */


#include "server.h"

extern struct misc_meta_data *mmd;
static FILE *audio_file = NULL;
static char supported_audio_formats[MAXLINE];

#if 1
	int mp3_init(void *);
#endif

#ifdef HAVE_OGGVORBIS
	int ogg_init(void *);
#endif

static struct audio_format afl[] = { /* audio format list */
#if 1
	{
		.name = "mp3",
		.init = mp3_init
	},
#endif
#ifdef HAVE_OGGVORBIS
	{
		.name = "ogg",
		.init = ogg_init
	},
#endif
	{
		.name = NULL,
	}
};

const char *audio_format_name(int i)
{
	return i >= 0?  afl[i].name : "(none)";
}

char *afs_stream_writer_cmd(int i)
{
	return i >=0? afl[i].open_cmd : "/bin/false";
}


char *afs_supported_audio_formats(void)
{
	return supported_audio_formats;
}

static FILE *afs_open_stream_writer(int format_num)
{
	struct audio_format *af;
	long flags;

	if (format_num < 0)
		return NULL;
	af = afl + format_num;

	clog(INFO, "%s: opening %s stream_writer\n", __func__, af->name);
	af->sw = popen_write_cmdline_pid(&af->sw_pid, af->open_cmd);
	
	//sleep(1);
	if (!af->sw)
		clog(ERROR, "%s: failed to open %s stream writer\n", __func__,
			af->name);
	else {
		clog(DEBUG, "%s: %s stream writer pid: %d, fd: %d\n", __func__, 
			af->name, af->sw_pid, fileno(af->sw));
		/* set non-blocking io */
		flags = fcntl(fileno(af->sw), F_GETFL) | O_NONBLOCK;
		fcntl(fileno(af->sw), F_SETFL, flags);
	}
	return af->sw;
}
static void setup_stream_command(struct gengetopt_args_info *conf, int num)
{
	int i;
	char *p, *q;

	if (num < 0)
		return;
	for (i = 0; i < conf->stream_write_cmd_given; ++i) {
		q = strdup(conf->stream_write_cmd_arg[i]);
		p = strchr(q, ':');
		if (!p) {
			clog(WARNING, "%s invalid stream_write_cmd\n", __func__);
			goto next;
		}
		*p = '\0';
		p++;
		if (strcmp(q, afl[num].name))
			goto next;
		afl[num].open_cmd = conf->stream_write_cmd_arg[i] + strlen(q) + 1;
		clog(INFO, "%s: using \"%s\" for files of type %s\n", 
			__func__, afl[num].open_cmd, afl[num].name);
next:
		free(q);
	}
}
int afs_init(struct gengetopt_args_info *conf)
{
	int i;

	clog(INFO, "%s: initializing audio file sender\n", __func__);
	*supported_audio_formats = '\0';
	for (i = 0; afl[i].name; i++) {
		clog(DEBUG, "%s: supported audio format: %s\n", __func__, 
			afl[i].name);
		if (afl[i].init(&afl[i]) < 0) {
			clog(ERROR, "%s: init %s failed, disabling "
				"%s support\n", __func__, afl[i].name, 
				afl[i].name);
			afl[i].get_file_info = NULL;
			continue;
		}
		setup_stream_command(conf, i);
		sprintf(supported_audio_formats, "%s%s ", 
			*supported_audio_formats? supported_audio_formats: "",
			 afl[i].name);
		afl[i].sw_pid = 0;
		afl[i].sw = NULL;
	}
	if (!i) {
		clog(EMERG, "%s: no audio formats supported\n");
		exit(EXIT_FAILURE);
	}
	clog(DEBUG, "%s: supported audio formats: %s\n", __func__, 
			supported_audio_formats);
	
	return i;
}

/*
 * compute human readable string containing
 * afs_status for given integer value
 */
void afs_status_tohuman(unsigned int flags, char *msg)
{
	char *txt;

	if (flags & AFS_PLAYING)
		txt = "playing";
	else if (flags & AFS_NEXT)
		txt = "stopped";
	else
		txt = "paused";
	sprintf(msg, "%s", txt);
}

void *afs_get_status_flags(unsigned int flags, char *msg)
{
	msg[0]	= (flags & AFS_PLAYING)? 'P' : '_';
	msg[1] = (flags & AFS_NOMORE)? 'O' : '_';
	msg[2] = (flags & AFS_NEXT)? 'N' : '_';
	msg[3] = (flags & AFS_REPOS)? 'R' : '_';
	msg[4] = '\0';
	return msg;
}

int update_dbinfo(void)
{
	FILE *pipe;
	char string[STRINGSIZE];
	char *info_cmd[] = {"info", mmd->filename, NULL};
	int i;
	char prefix[MAXLINE];

	mmd->afs_status_flags &= ~AFS_REREAD_DB_INFO;
	clog(DEBUG, "%s: getting dbtool info %s\n", __func__, mmd->filename);
	if (!(pipe = popen_read_dbtool(info_cmd))) {
		clog(ERROR, "%s: popen_read_dbtool failed\n", __func__);
		return -1;
	}

	i = 0; /* number of dbinfo lines, maximum: 3 */
	mmd->dbinfo[0] = '\0';
	while (i < 3 && strlen(mmd->dbinfo) + MAXLINE + 10 < MMD_INFO_SIZE
			&& fgets(string, MAXLINE, pipe)) {
		sprintf(prefix, "dbinfo%d:", ++i);
		strcat(mmd->dbinfo, prefix);
		strcat(mmd->dbinfo, string);
		clog(INFO, "%s: dbinfo#%i: %s", __func__, i, string);
	}
	clog(DEBUG, "%s: read %i lines (%i bytes) of dbinfo\n", __func__,
		i, strlen(mmd->dbinfo));
	/* output always three lines */
	while (i < 3 && strlen(mmd->dbinfo) + 10 < MMD_INFO_SIZE) {
		clog(DEBUG, "%s: padding dbinfo\n", __func__);
		sprintf(prefix, "dbinfo%d:\n", ++i);
		strcat(mmd->dbinfo, prefix);
	}
	pclose(pipe);
	return 1;
}

static int get_file_info(int i)
{
	int ret = afl[i].get_file_info(audio_file, mmd->audio_file_info,
		&mmd->chunks_total, &mmd->seconds_total);
	if (ret > 0)
		clog(INFO, "%s: valid %s file\n", __func__, afl[i].name);
	else
		clog(DEBUG, "%s: %s format not recognized\n", __func__, 
			afl[i].name);
	return ret;
}

static int guess_audio_format(const char *name)
{

	int i, len1 = strlen(name), len2;

	for (i = 0; afl[i].name; i++) {
		len2 = strlen(afl[i].name);
		if (len1 < len2)
			continue;
		if (!strncasecmp(name + (len1 - len2), afl[i].name, len2)) {
			clog(DEBUG, "%s: might be %s\n", __func__,
				afl[i].name);
			return i;
		}
	}
	return -1;
}

static int get_audio_format(int omit)
{
	int i;

	for (i = 0; afl[i].name; i++) {
		if (i == omit || !afl[i].get_file_info)
			continue;
		rewind(audio_file);
		if (get_file_info(i) > 0)
			return i;
		rewind(audio_file);
	}
	return -1;
}

/*
 * upddate shared mem
 */
static int update_mmd(void)
{
	FILE *pipe;
	char *us_cmd[] = {"us", mmd->filename, NULL};
	int i;
	struct stat file_status;

	clog(INFO, "%s: current stream: %s, num_played: %u\n", __func__,
		mmd->stream_name, mmd->num_played);
	i = guess_audio_format(mmd->filename);
	if (i < 0 || get_file_info(i) < 0)
		i = get_audio_format(i);
	if (i < 0) {
		clog(INFO, "%s: unknown audio format\n", __func__,
			afl[i].name, __func__);
		return -1;
	}
	mmd->audio_format = i;
	mmd->chunks_sent = 0;
	if (fstat(fileno(audio_file), &file_status) == -1) {
		clog(ERROR, "%s: fstat error (%s)\n", __func__, strerror(errno));
		fclose(audio_file);
		return -1;
	}
	mmd->size = file_status.st_size;
	strncpy(mmd->mtime, ctime(&file_status.st_mtime), MAXLINE);
	clog(INFO, "%s: size: %lu, mtime: %s", __func__, mmd->size, mmd->mtime);
	if (update_dbinfo() < 0)
		return -1;

	clog(DEBUG, "%s: popen para_dbtool us\n", __func__);
	if (!(pipe = popen_read_dbtool(us_cmd))) {
		clog(ERROR, "%s: popen_read_dbtool failed\n", __func__);
		return -1;
	}
	pclose(pipe);
	strcpy(mmd->old_stream_name, mmd->stream_name);
	clog(NOTICE, "%s: next song: %s\n", __func__, mmd->filename);
	return 1;
}

/*
 * get next song via external command dbtool sl. Returns -1 if no valid song
 * could be found, Otherwise, audio_file contains valid file and 1 is
 * returned.
 */
static void get_song(void)
{
	FILE *pipe;
	char buf[MAXLINE];
	int found_file = 0;
	char *strptr;
	char *cmd[] = {"sl", "100", mmd->stream_name, NULL};

	clog(DEBUG, "%s: Getting next song from stream %s\n", __func__,
			mmd->stream_name);
	if (!(pipe = popen_read_dbtool(cmd))) {
		clog(ERROR, "%s: popen_read_dbtool failed\n", __func__);
		goto out;
	}
	while ((fgets(buf, MAXLINE - 1, pipe))) {
		char *name;
		char tmp[MAXLINE] = "";
		int pos = 0;

		/* cut trailing \n */
		if ((strptr = strchr(buf, '\n')))
			*strptr = '\0';
		sscanf(buf, "%d%n%s", &mmd->score, &pos, tmp);
		if (!pos || !tmp[0])
			continue;
		name = buf + pos + 1;
		clog(DEBUG, "%s: Opening %s\n", __func__, name);
		if (!(audio_file = fopen(name, "r"))) {
			clog(DEBUG, "%s: Can't open %s\n", __func__, name);
			continue;
		}
		strcpy(mmd->filename, name);
		if (update_mmd() < 0)
			continue;
		mmd->num_played++;
		found_file = 1;
		break;
	}
out:	
	if (pipe)
		pclose(pipe);
	if (!found_file) {
		clog(ERROR, "%s: dbtool didn't find any valid file\n", __func__);
		mmd->afs_status_flags = AFS_NEXT;
		strcpy(mmd->old_stream_name, mmd->stream_name);
		strcpy(mmd->stream_name, "(none)");
	} else 
		clog(DEBUG, "%s: Success (fd = %d)\n", __func__, 
			fileno(audio_file));
}

/*
 * called from main when playing, just before main calls select
 */
FILE *afs_preselect(void)
{
	int flags = mmd->afs_status_flags;
	struct audio_format *af = afl + mmd->audio_format;

	mmd->afs_status_flags &= (~(AFS_NEXT|AFS_REPOS));

	if (flags & AFS_NEXT && audio_file && af->sw) {
		af->end_of_input(audio_file, &af->sw, &af->sw_pid);
		//fclose(audio_file);
		audio_file = NULL;
		return NULL;
	}
	if (!audio_file) {
		clog(DEBUG, "%s: no audio file\n", __func__);
		get_song();
		goto check_sw;
	}
	if ((flags & AFS_REPOS)) {
		clog(DEBUG, "%s: calling reposition handler (%lu)\n", 
			__func__, mmd->chunks_sent);
//		stop_stream(audio_file);
		af->reposition_stream(audio_file, mmd->chunks_sent, 
			&af->sw_pid);
		if (af->sw && !af->sw_pid) {
			pclose(af->sw);
			af->sw = NULL;
		}
	}
check_sw:
	//clog(DEBUG, "%s: checking sw\n", __func__); 
	af = afl + mmd->audio_format;
	if (!(mmd->afs_status_flags & AFS_PLAYING))
		return NULL;
	if (!af->sw) {
		clog(DEBUG, "%s: open stream\n", __func__); 
		af->sw = afs_open_stream_writer(mmd->audio_format);
	}
	//clog(DEBUG, "%s: returning %d\n", __func__, af->sw? fileno(af->sw) : 42); 
	return af->sw;
}

/*
 * Status is playing and stream_writer_pipe is ready for writing. So we read
 * one chunk of data and send it to pipe.
 */
int afs_send_chunk(FILE *stream_writer)
{
	int size, ret;
	struct audio_format *af = &afl[mmd->audio_format];

	if (mmd->audio_format < 0)
		return -1;

	if (mmd->afs_status_flags & AFS_NEXT 
			|| mmd->afs_status_flags & AFS_REPOS)
		return 0;
	if (!audio_file) {
		clog(CRIT, "%s: BUG: no audio file\n", __func__);
		return -1;
	}
	//clog(DEBUG, "%s: \n", __func__);
	ret = af->read_chunk(audio_file);
	if (ret < 0) {
		clog(ERROR, "%s: read error, skipping\n", __func__);
		goto err_out;
	}
	if (!ret) {
		clog(DEBUG, "%s: song finished (flags:%x)\n", __func__,
				mmd->afs_status_flags);
		goto err_out;
	}
	size = ret;
	ret = af->write_chunk(stream_writer, mmd->audio_file_info, 
		&mmd->chunks_sent);
	if (ret < 0) {
		clog(ERROR, "%s: write error, skipping\n", __func__);
		goto err_out;
	}
	mmd->afs_status_flags |= AFS_PLAYING;
	return 1;
err_out:
	af->end_of_input(audio_file, &af->sw, &af->sw_pid);
	audio_file = NULL;
	//pclose(af->sw);
	//af->sw_pid = 0;
	//af->sw = NULL;
	if ((mmd->afs_status_flags & AFS_NOMORE))
		mmd->afs_status_flags = AFS_NEXT;
	else
		get_song();
	return ret;
}
